from logging import Logger

import torch
from torch import Tensor
from torch.utils.data import Dataset


class PointDataset(Dataset):
    def __init__(self, samples: Tensor, evaluations: Tensor):
        super().__init__()
        self.samples = samples
        self.evaluations = evaluations

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx):
        return self.samples[idx], self.evaluations[idx]


class TuplesDataset(Dataset):
    def __init__(
        self, i_indexes: Tensor, j_indexes: Tensor, database: Tensor, values: Tensor
    ):
        super().__init__()
        self.i_indexes = i_indexes
        self.j_indexes = j_indexes
        self.database = database.detach()
        self.values = values.detach()

    def __len__(self):
        return len(self.i_indexes)

    def __getitem__(self, idx):
        j_index = self.j_indexes[idx]
        i_idx = self.i_indexes[idx]
        x_i = self.database[i_idx].detach()
        x_j = self.database[j_index].detach()
        y_i = self.values[i_idx].detach()
        y_j = self.values[j_index].detach()
        return x_i, x_j, y_i, y_j


class SinglePairPerPointDataset(TuplesDataset):
    def __init__(self, database: Tensor, values: Tensor, exploration_size: int):
        len_replay_buffer = len(values)
        i_indexes = torch.randperm(len_replay_buffer, device=database.device)

        i_reference = torch.randint(
            0, exploration_size, size=(len_replay_buffer,), device=database.device
        )
        # We are trying to batch each sample with his corresponding samples,
        # so we refer from num of exploration samples
        explore_indexes = torch.div(i_indexes, exploration_size, rounding_mode="trunc")

        j_indexes = exploration_size * explore_indexes + i_reference
        super().__init__(i_indexes, j_indexes, database, values)


class PairsInRangeDataset(TuplesDataset):
    def __init__(self, database: Tensor, values: Tensor, exploration_size: int):
        len_replay_buffer = len(values)
        single_i_indexes = torch.randperm(len_replay_buffer, device=database.device)
        i_indexes = single_i_indexes.repeat(exploration_size)

        i_reference_single = torch.randint(
            0, exploration_size, size=(len_replay_buffer,), device=database.device
        )
        i_reference = torch.cat(
            [
                (i_reference_single + i) % exploration_size
                for i in range(exploration_size)
            ],
            axis=0,
        )
        # We are trying to batch each sample with his corresponding samples,
        # so we refer from num of exploration samples
        explore_indexes = torch.div(i_indexes, exploration_size, rounding_mode="trunc")

        j_indexes = exploration_size * explore_indexes + i_reference
        non_same_samples = i_indexes != j_indexes
        super().__init__(
            i_indexes[non_same_samples], j_indexes[non_same_samples], database, values
        )


class PairsInEpsRangeDataset(TuplesDataset):
    def __init__(
        self,
        database: Tensor,
        values: Tensor,
        epsilon: float,
        max_tuples: int = None,
        logger: Logger = None,
    ):
        database = database.detach()
        distances = torch.cdist(database, database)

        only_upper_half_matrix = torch.triu(
            torch.ones_like(distances, device=database.device, dtype=torch.bool),
            diagonal=1,
        )
        close_to_epsilon = distances < epsilon
        distances[~(close_to_epsilon & only_upper_half_matrix)] = 0
        tuples = distances.nonzero()
        self.size = max_tuples if max_tuples else len(tuples)
        if len(tuples) == 0:
            logger.error(
                f"No tuples, min distance: {distances.min()}, max distance: {distances.max()} epsilon {epsilon}"
            )
            super().__init__(
                torch.tensor([], device=database.device),
                torch.tensor([], device=database.device),
                database,
                values,
            )
        else:
            tuples_idx = torch.randint(len(tuples), (self.size,))
            super().__init__(
                tuples[tuples_idx, 0], tuples[tuples_idx, 1], database, values
            )

    def __len__(self):
        return self.size


class NewPairEpsDataset(Dataset):
    def __init__(
        self,
        database: Tensor,
        values: Tensor,
        epsilon: float,
        new_samples: Tensor,
    ):
        super().__init__()
        self.database = database.detach()
        self.new_samples = new_samples
        self.values = values
        distances = torch.cdist(new_samples, database)

        only_upper_half_matrix = torch.triu(
            torch.ones_like(distances, device=database.device, dtype=torch.bool),
            diagonal=1,
        )
        close_to_epsilon = distances < epsilon
        distances[~(close_to_epsilon & only_upper_half_matrix)] = 0
        self.tuples = distances.nonzero()

    def __len__(self):
        return len(self.tuples)

    def __getitem__(self, idx):
        j_index = self.tuples[idx, 0]
        i_idx = self.tuples[idx, 1]
        x_i = self.database[i_idx].detach()
        x_j = self.new_samples[j_index].detach()
        y_i = self.values[i_idx].detach()
        y_j = self.values[j_index].detach()
        return x_i, x_j, y_i, y_j


class PairFromDistributionDataset(TuplesDataset):
    def __init__(self, dataset: TuplesDataset):
        values = dataset.values.detach()
        values_diff = (values[dataset.i_indexes] - values[dataset.j_indexes]).abs()
        values_diff_distributions = values_diff / values_diff.sum()
        distributed_indices = torch.multinomial(
            values_diff_distributions, len(values_diff)
        )
        super().__init__(
            i_indexes=dataset.i_indexes[distributed_indices],
            j_indexes=dataset.j_indexes[distributed_indices],
            database=dataset.database,
            values=dataset.values,
        )



from typing import Callable, Type, Tuple

import torch
from torch import Tensor
from torch.nn import Module
from torch.optim import Optimizer
from torch.utils.data import Dataset
from torch.utils.data import WeightedRandomSampler

from .convergence import ConvergenceAlgorithm
from .distribution import WeightsDistributionBase
from .losses import loss_with_quantile
from .trainer import train_gradient_network, step_model_with_gradient
from .utils import reset_all_weights, loss_from_taylor_loss


class EGL(ConvergenceAlgorithm):
    def __init__(
        self,
        *args,
        gradient_network: Module,
        gradient_optimizer: Optimizer,
        grad_loss: Callable,
        num_of_minibatch_to_train: int,
        database_type: Type[Dataset],
        dataset_parameters: Callable,
        weight_func: WeightsDistributionBase = None,
        taylor_loss: Callable[
            [Tensor, Tensor, Tensor, Tensor], Tuple[Tensor, Tensor]
        ] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.grad_network = gradient_network
        self.gradient_optimizer = gradient_optimizer
        self.grad_loss = grad_loss
        self.dataset_parameters = dataset_parameters
        self.num_of_minibatch_to_train = num_of_minibatch_to_train
        self.database_type = database_type
        self.weight_func = weight_func
        self.taylor_loss = taylor_loss

    def train_surrogate(self, samples: Tensor, samples_value: Tensor, batch_size: int):
        self.grad_network.train()
        mapped_evaluations = self.value_normalizer.map(samples_value)
        additional_params = self.dataset_parameters(self)
        dataset = self.database_type(
            database=samples,
            values=mapped_evaluations,
            **additional_params,
            logger=self.logger,
        )

        if self.weight_func:
            self.weight_func.pre_training(dataset.database, dataset.values)

        loss = loss_from_taylor_loss(self.taylor_loss, self.grad_loss)
        sampler = None
        if self.weight_func:
            distribute_weights = self.weight_func.distribution_from_dataset(dataset)
            if distribute_weights.device.type == "mps":
                distribute_weights = distribute_weights.cpu()
            sampler = WeightedRandomSampler(
                distribute_weights.detach().clone(), len(dataset)
            )

        train_gradient_network(
            loss,
            self.gradient_optimizer,
            dataset,
            batch_size,
            self.logger,
            # self.num_of_minibatch_to_train,
            sampler=sampler,
        )
        self.grad_network.eval()

    def calc_loss(self, value: Tensor, target: Tensor) -> Tensor:
        if self.weight_func:
            return loss_with_quantile(value, target, self.weight_func, self.grad_loss)
        return self.grad_loss(value, target)

    def train_model(self):
        self.gradient_optimizer.zero_grad()
        self.grad_network.eval()

        curr_point_gradient = self.gradient(self.curr_point)
        self.logger.info(
            f"Algorithm {self.__class__.__name__} "
            f"moving Gradient size: {torch.norm(curr_point_gradient)} on {self.env}"
        )
        step_model_with_gradient(
            self.curr_point, curr_point_gradient, self.function_opt
        )

    def gradient(self, x) -> Tensor:
        training = self.grad_network.training
        self.grad_network.eval()
        model_to_train_gradient = self.grad_network(x.unsqueeze(0)).squeeze(0)
        self.grad_network.train(training)
        model_to_train_gradient[model_to_train_gradient != model_to_train_gradient] = 0
        return model_to_train_gradient

    def explore(self, exploration_size: int):
        samples, evaluations = super().explore(exploration_size)
        if self.weight_func:
            self.weight_func.update(samples=samples, values=evaluations)
        return samples, evaluations

    def after_shrinking_hook(self):
        reset_all_weights(self.grad_network)
        if self.weight_func:
            self.weight_func.restart()



from typing import Protocol

from torch import Tensor

from .exception import NoMoreBudgetError


class Function(Protocol):
    def evaluate(self, data: Tensor) -> Tensor:
        raise NotImplementedError()

    def denormalize(self, data: Tensor) -> Tensor:
        """
        map data from (-1,1) to entire space
        """
        raise NotImplementedError()

    def normalize(self, data: Tensor) -> Tensor:
        raise NotImplementedError()


class BasicFunction(Function):
    def __init__(self, func, budget=150_000, lower_bound=-5, upper_bound=5):
        self.func = func
        self.curr_budget = 0
        self.budget = budget
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def evaluate(self, data: Tensor) -> Tensor:
        self.curr_budget += len(data)
        if self.curr_budget > self.budget:
            raise NoMoreBudgetError(f"Exceeded budget of {self.curr_budget}")
        if data.ndim == 1:
            data = data.unsqueeze(0)
            return self.func(data).squeeze(0)
        return self.func(data)

    def denormalize(self, data: Tensor) -> Tensor:
        return (
            0.5 * (data + 1) * (self.upper_bound - self.lower_bound) + self.lower_bound
        )

    def normalize(self, data: Tensor) -> Tensor:
        return (
            (data - self.lower_bound) / (self.upper_bound - self.lower_bound) * 2
        ) - 1



import abc
from abc import ABC

import torch
from torch import Tensor

TANH_UPPER_LIMIT = 10
TANH_LOWER_LIMIT = -10


class TrustRegion(ABC):
    @abc.abstractmethod
    def map(self, tensor: Tensor) -> Tensor:
        raise NotImplementedError()

    @abc.abstractmethod
    def inverse(self, tensor: Tensor) -> Tensor:
        raise NotImplementedError()

    @abc.abstractmethod
    def squeeze(self, best_result: Tensor):
        raise NotImplementedError()

    @abc.abstractmethod
    def move_center(self, best_result: Tensor):
        raise NotImplementedError()

    @abc.abstractmethod
    def unsqueeze(self):
        raise NotImplementedError()


class StaticShrinkingTrustRegion(TrustRegion, ABC):
    def __init__(
        self,
        lower_bounds: Tensor,  # NOTE: This is legacy - Should be deleted for package
        upper_bounds: Tensor,
        shrink_factor: float = 0.9,
        min_sigma: float = 0.1,
        min_trust_region_size: float = 1e-13,
        dim_proportion_factor: float = 0,
        dtype: torch.dtype = torch.float64,
        device: int = None,
    ):
        shape = lower_bounds.shape
        self.lower_bounds = lower_bounds.to(device=device)
        self.upper_bounds = upper_bounds.to(device=device)
        self.min_trust_region_size = min_trust_region_size
        self.mu = torch.zeros(shape, device=device, dtype=dtype)
        self.sigma = torch.ones(shape, device=device, dtype=dtype)
        self.dim_proportion_factor = dim_proportion_factor
        self.shrink_factor = shrink_factor
        self.min_sigma = min_sigma * torch.ones(shape, device=device, dtype=dtype)
        self.dtype = dtype

    @abc.abstractmethod
    def normalize_to_real(self, data: Tensor) -> Tensor:
        raise NotImplementedError()

    @abc.abstractmethod
    def normalize_to_unreal(self, data: Tensor) -> Tensor:
        raise NotImplementedError()

    @property
    def device(self):
        return self.mu.device

    def _move_to_best_device(self, tensor: Tensor):
        tensor_device = tensor.device
        if tensor_device == torch.device("cpu"):
            tensor = tensor.to(device=self.device)
            mu, sigma = self.mu, self.sigma
        else:
            mu, sigma = (
                self.mu.to(device=tensor_device),
                self.sigma.to(device=tensor_device),
            )
        return tensor, mu, sigma

    def map(self, tensor) -> Tensor:
        original_shape = tensor.shape
        tensor, mu, sigma = self._move_to_best_device(tensor)
        tensor = (tensor - mu) / sigma.view(1, -1)
        tensor = self.normalize_to_unreal(tensor)
        return tensor.reshape(original_shape)

    def inverse(self, tensor) -> Tensor:
        tensor, mu, sigma = self._move_to_best_device(tensor)
        normalized_tensor = self.normalize_to_real(tensor)
        tensor = mu + sigma * normalized_tensor
        tensor = torch.clamp(tensor, min=-1, max=1)
        return tensor

    def squeeze(self, best_result):
        best_result = best_result.to(device=self.device)
        # max_dimensional_delta = gradient.abs().max()
        shrink_mask = (best_result < self.mu + (1 - self.min_sigma) * self.sigma) | (
            best_result > self.mu - (1 - self.min_sigma) * self.sigma
        )
        # gradient_mask = gradient.abs() > max_dimensional_delta / self.dim_proportion_factor
        if self.dim_proportion_factor:
            shrink_mask = shrink_mask  # & gradient_mask
        self.sigma[shrink_mask] = self.shrink_factor * self.sigma[shrink_mask]
        self.mu = best_result.detach()

    def move_center(self, best_result: Tensor):
        self.mu = best_result.detach()

    def unsqueeze(self):
        self.sigma = self.sigma / self.shrink_factor


class TanhTrustRegion(StaticShrinkingTrustRegion):
    def normalize_to_real(self, data: Tensor) -> Tensor:
        return torch.tanh(data)

    def normalize_to_unreal(self, data: Tensor) -> Tensor:
        data = torch.clamp(data, min=-1 + 1e-16, max=1 - 1e-16)
        return torch.arctanh(data)


class LinearTrustRegion(StaticShrinkingTrustRegion):
    def normalize_to_real(self, data: Tensor) -> Tensor:
        normalized_tensor = (
            ((data - self.lower_bounds) / (self.upper_bounds - self.lower_bounds)) * 2
        ) - 1
        return torch.clip(normalized_tensor, min=-1, max=1)

    def normalize_to_unreal(self, data: Tensor) -> Tensor:
        return (
            0.5 * (data + 1) * (self.upper_bounds - self.lower_bounds)
            + self.lower_bounds
        )



from abc import ABC


class AlgorithmCallbackHandler(ABC):
    def on_algorithm_start(self, alg, *args, **kwargs):
        raise NotImplementedError()

    def on_epoch_end(self, alg, *args, **kwargs):
        raise NotImplementedError()

    def on_algorithm_update(self, alg, *args, **kwargs):
        raise NotImplementedError()

    def on_algorithm_end(self, alg, *args, **kwargs):
        raise NotImplementedError()


class CallableForEpochEnd(AlgorithmCallbackHandler):
    def __init__(self, callback):
        self.callback = callback

    def on_algorithm_start(self, alg, *args, **kwargs):
        pass

    def on_epoch_end(self, alg, *args, **kwargs):
        if self.callback:
            self.callback(alg, *args, **kwargs)

    def on_algorithm_update(self, alg, *args, **kwargs):
        pass

    def on_algorithm_end(self, alg, *args, **kwargs):
        pass



import abc
from abc import ABC

import numpy as np
import torch
from cma import CMAEvolutionStrategy
from egl.datsets import TuplesDataset
from torch import Tensor
from torch.distributions import MultivariateNormal


def prepare_batches_calc(dataset: TuplesDataset, max_memory_mb: int):
    # Batched implementation to respect memory constraints
    total_size = len(dataset.i_indexes)
    dim_size = dataset.database.shape[1]
    device = dataset.database.device
    dtype = dataset.database.dtype

    # Estimate memory per sample (rough approximation)
    # Each sample requires x_i, x_j (2 * dim_size floats)
    estimated_mem_per_sample = (
        2 * dim_size * torch.tensor([], dtype=dtype).element_size()
    )

    # Calculate batch size based on max_memory
    max_mem_bytes = max_memory_mb * (1024**2)
    batch_size = max(1, int(max_mem_bytes / estimated_mem_per_sample))

    # Initialize empty weights tensor
    all_weights = torch.zeros(total_size, device=device, dtype=dtype)
    return all_weights, batch_size, total_size


class WeightsDistributionBase(ABC):
    def __init__(self, max_memory_usage: int = None):
        self.max_memory_usage = max_memory_usage

    def update(self, *args, **kwargs):
        pass

    def restart(self):
        pass

    def pre_training(self, samples: Tensor, values: Tensor):
        pass

    @abc.abstractmethod
    def distribute_weights(
        self, x_i: Tensor, x_j: Tensor, y_i: Tensor, y_j: Tensor
    ) -> Tensor:
        raise NotImplementedError()

    def _weights_for_batch(self, x_i: Tensor, x_j: Tensor, y_i: Tensor, y_j: Tensor):
        return self.distribute_weights(x_i, x_j, y_i, y_j)

    def _post_process_weights(self, weights: Tensor):
        return weights

    def distribution_from_dataset(self, dataset: TuplesDataset):
        if self.max_memory_usage is None:
            # Original implementation - process all at once
            x_i = dataset.database[dataset.i_indexes]
            x_j = dataset.database[dataset.j_indexes]
            y_i = dataset.values[dataset.i_indexes]
            y_j = dataset.values[dataset.j_indexes]
            weights = self.distribute_weights(x_i, x_j, y_i, y_j).detach().clone()
            return weights
        else:
            all_weights, batch_size, total_size = prepare_batches_calc(
                dataset, self.max_memory_usage
            )

            # Process in batches
            for start_idx in range(0, total_size, batch_size):
                end_idx = min(start_idx + batch_size, total_size)

                # Get batch indices
                batch_i_indexes = dataset.i_indexes[start_idx:end_idx]
                batch_j_indexes = dataset.j_indexes[start_idx:end_idx]

                # Extract batch data
                x_i = dataset.database[batch_i_indexes]
                x_j = dataset.database[batch_j_indexes]
                y_i = dataset.values[batch_i_indexes]
                y_j = dataset.values[batch_j_indexes]

                # Process batch
                batch_weights = self._weights_for_batch(x_i, x_j, y_i, y_j).detach()

                # Store results
                all_weights[start_idx:end_idx] = batch_weights

            final_weights = self._post_process_weights(all_weights)
            return final_weights


class QuantileWeights(WeightsDistributionBase):
    def __init__(self, train_quantile: int = 50, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.train_quantile = train_quantile
        self.quantile_value = None

    def pre_training(self, samples: Tensor, values: Tensor):
        self.quantile_value = torch.quantile(values, self.train_quantile / 100)

    def distribute_weights(
        self, x_i: Tensor, x_j: Tensor, y_i: Tensor, y_j: Tensor
    ) -> Tensor:
        data = torch.min(y_i, y_j)
        indices = data < self.quantile_value
        weights = torch.zeros_like(data, dtype=data.dtype)
        weights[indices] = 1.0
        return weights.detach()


class CMAWeights(WeightsDistributionBase):
    def __init__(self, dims: int, device: int, dtype: torch.dtype, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dims = dims
        self.device = device
        self.dtype = dtype
        self.cma = None
        self.mvn = None
        self.sum_probs = None
        self.database_size = None
        self.restart()

    def update(self, *args, samples: Tensor, values: Tensor, **kwargs):
        self.cma.ask()
        self.cma.tell(samples.detach().cpu().numpy(), values.detach().cpu().numpy())
        self.mvn_from_cma()

    def restart(self):
        self.cma = CMAEvolutionStrategy(np.array([0.0] * self.dims), 0.5)
        self.mvn_from_cma()

    def mvn_from_cma(self):
        covariance_matrix = torch.from_numpy(self.cma.sm.covariance_matrix).to(
            device=self.device, dtype=self.dtype
        )
        # cov_sym = (covariance_matrix + covariance_matrix.T) / 2
        cov_sym = covariance_matrix
        w, v = torch.linalg.eigh(cov_sym)
        w_clipped = w.clamp(min=1e-6)
        c_psd = v @ w_clipped.diag() @ v.T
        c_psd = (c_psd + c_psd.T) / 2

        mean = torch.from_numpy(self.cma.mean).to(device=self.device, dtype=self.dtype)
        self.mvn = MultivariateNormal(mean, c_psd)

    def pre_training(self, samples: Tensor, values: Tensor):
        log_probs = self.mvn.log_prob(samples)
        probs = torch.exp(log_probs)
        probs[probs.isnan()] = 0
        self.sum_probs = probs.sum()
        self.database_size = len(samples)

    def distribute_weights(
        self, x_i: Tensor, x_j: Tensor, y_i: Tensor, y_j: Tensor
    ) -> Tensor:
        points = torch.where((y_i < y_j).unsqueeze(1), x_i, x_j).detach()
        log_probs = self.mvn.log_prob(points)
        probs = torch.exp(log_probs)
        probs[probs.isnan()] = 0
        multiplier = self.database_size // 10
        probability = probs * multiplier / (self.sum_probs.clip(min=1e-17))
        return probability.detach()


class CMAWeightsPerSample(CMAWeights):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.samples = None
        self.values = None

    def pre_training(self, samples: Tensor, values: Tensor):
        self.samples = samples
        self.values = values
        self.restart()

    def distribute_weights(
        self, x_i: Tensor, x_j: Tensor, y_i: Tensor, y_j: Tensor
    ) -> Tensor:
        a_exp = self.samples.unsqueeze(0)
        b_exp = x_i.unsqueeze(1)
        matches = (a_exp == b_exp).all(dim=2)
        indexes = torch.where(
            matches, torch.arange(a_exp.size(0)).to(device=matches.device), -1
        )
        indexes = torch.max(indexes, dim=1).values
        x_i_values = y_i[indexes]
        self.update(samples=x_i, values=x_i_values)

        b_exp = x_i.unsqueeze(1)
        matches = (a_exp == b_exp).all(dim=2)
        indexes = torch.where(
            matches, torch.arange(a_exp.size(0)).to(device=matches.device), -1
        )
        indexes = torch.max(indexes, dim=1).values
        x_j_values = y_j[indexes]
        self.update(samples=x_j, values=x_j_values)
        self.mvn_from_cma()
        super().pre_training(samples=x_i, values=y_i)
        return super().distribute_weights(x_i, x_j, y_i, y_j)


class WeightDistributionNormalization(WeightsDistributionBase):
    def __init__(self, weights_creator: WeightsDistributionBase, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.weights_creator = weights_creator

    def distribute_weights(
        self, x_i: Tensor, x_j: Tensor, y_i: Tensor, y_j: Tensor
    ) -> Tensor:
        weights = self.weights_creator.distribute_weights(x_i, x_j, y_i, y_j)
        weights_sum = weights.sum()
        if weights_sum == 0:
            weights = torch.ones_like(weights)
            weights_sum = weights.sum()
        return (weights / weights_sum).detach()


class WeightSmoother(WeightsDistributionBase):
    def __init__(
        self,
        inner_weights_creator: WeightsDistributionBase,
        smooth_alpha: float = 0.4,
        smooth_epsilon: float = 1e-4,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.inner_weights_creator = inner_weights_creator
        self.smooth_alpha = smooth_alpha
        self.smooth_epsilon = smooth_epsilon

    def update(self, *args, **kwargs):
        self.inner_weights_creator.update(*args, **kwargs)

    def restart(self):
        self.inner_weights_creator.restart()

    def pre_training(self, samples: Tensor, values: Tensor):
        self.inner_weights_creator.pre_training(samples=samples, values=values)

    def distribute_weights(
        self, x_i: Tensor, x_j: Tensor, y_i: Tensor, y_j: Tensor
    ) -> Tensor:
        weights = self.inner_weights_creator.distribute_weights(x_i, x_j, y_i, y_j)
        normalized_weights = (weights + self.smooth_epsilon).pow(self.smooth_alpha)
        final_weights = normalized_weights / normalized_weights.sum()
        final_weights[final_weights.isnan()] = 0
        return final_weights.detach()

    def _post_process_weights(self, weights: Tensor):
        normalized_weights = (weights + self.smooth_epsilon).pow(self.smooth_alpha)
        final_weights = normalized_weights / normalized_weights.sum()
        final_weights[final_weights.isnan()] = 0
        return final_weights

    def _weights_for_batch(self, x_i: Tensor, x_j: Tensor, y_i: Tensor, y_j: Tensor):
        weights = self.inner_weights_creator.distribute_weights(x_i, x_j, y_i, y_j)
        return weights.detach()



from logging import Logger
from typing import List, Tuple

from torch import Tensor
from torch.nn import Module
from torch.optim import Optimizer
from torch.utils.data import BatchSampler, RandomSampler, DataLoader, Sampler

from .datsets import TuplesDataset


def train_gradient_network(
    gradient_loss_func: Module,
    optimizer,
    dataset: TuplesDataset,
    batch_size: int,
    logger: Logger,
    sampler: Sampler = None,
) -> Tuple[List[float], List[float]]:
    losses = []

    sampler = sampler or RandomSampler(range(len(dataset)))
    sampler = BatchSampler(sampler, batch_size=batch_size, drop_last=False)
    data_loader = DataLoader(dataset, sampler=sampler)
    for i, (x_i, x_j, y_i, y_j) in enumerate(data_loader):
        optimizer.zero_grad()
        x_i, x_j, y_i, y_j = x_i[0], x_j[0], y_i[0], y_j[0]
        loss = gradient_loss_func(x_i, x_j, y_i, y_j)
        loss = loss.mean()
        loss.backward()
        optimizer.step()
        losses.append(loss.cpu().item())
    return losses


def step_model_with_gradient(point: Tensor, gradient: Tensor, optimizer: Optimizer):
    optimizer.zero_grad()
    point.grad = gradient
    optimizer.step()



import math

import torch
import torch.nn as nn
from torch.optim import Adam

from .datsets import PairsInEpsRangeDataset
from .distribution import CMAWeights
from .egl import EGL
from .function import BasicFunction
from .handlers import CallableForEpochEnd
from .losses import GradientLoss, NaturalHessianLoss
from .trust_region import TanhTrustRegion
from .value_normalizer import AdaptedOutputUnconstrainedMapping


def minimize(fun, x0, args=(), bounds=None, constraints=150_000, callback=None):
    use_hessian = args[0]
    taylor_loss_klass = NaturalHessianLoss if use_hessian else GradientLoss
    use_weights = args[1]
    dim = len(x0)

    if not isinstance(x0, nn.Parameter):
        x0 = nn.Parameter(x0.clone().detach())
    x0.requires_grad = True

    bounds = bounds or [-5, 5]
    func = BasicFunction(fun, constraints, *bounds)
    optimizer = Adam([x0], lr=0.01, eps=1e-04)
    device = x0.device
    gradient_network = nn.Sequential(
        nn.Linear(dim, 10),
        nn.ReLU(),
        nn.Linear(10, 15),
        nn.ReLU(),
        nn.Linear(15, dim),
    ).to(device=device, dtype=x0.dtype)
    grad_optimizer = Adam(gradient_network.parameters(), eps=1e-04)
    taylor_loss = taylor_loss_klass(gradient_network)
    num_of_minibatch_to_train = int(2000 * math.sqrt(dim))
    max_tuples = int(20_000 * math.floor(math.sqrt(dim)))
    egl = EGL(
        env=func,
        curr_point=x0,
        function_opt=optimizer,
        epsilon=0.8 * math.sqrt(dim),
        epsilon_factor=0.97,
        min_epsilon=1e-4,
        perturb=0,
        gradient_network=gradient_network,
        gradient_optimizer=grad_optimizer,
        grad_loss=nn.SmoothL1Loss(),
        num_of_minibatch_to_train=num_of_minibatch_to_train,
        database_type=PairsInEpsRangeDataset,
        dataset_parameters=lambda alg: {
            "epsilon": alg.epsilon,
            "max_tuples": max_tuples,
        },
        weight_func=CMAWeights(dim, device, x0.dtype) if use_weights else None,
        taylor_loss=taylor_loss,
        trust_region=TanhTrustRegion(
            lower_bounds=torch.tensor(
                [-func.lower_bound] * dim, device=device, dtype=x0.dtype
            ),
            upper_bounds=torch.tensor(
                [func.lower_bound] * dim, device=device, dtype=x0.dtype
            ),
        ),
        value_normalizer=AdaptedOutputUnconstrainedMapping(),
    )

    egl.train(
        epochs=constraints,
        exploration_size=int(8 * math.sqrt(dim)),
        num_loop_without_improvement=10,
        min_iteration_before_shrink=40,
        callback_handlers=[CallableForEpochEnd(callback)],
    )
    best_point = egl.best_model.detach().clone()
    return egl.env.denormalize(egl.trust_region.inverse(best_point))



import abc
import copy
import logging
import math
from logging import Logger
from typing import List

import torch
import torch.nn as nn
from torch import Tensor
from torch.optim import Optimizer
from tqdm.auto import trange

from .exception import NoMoreBudgetError, AlgorithmFinished
from .function import Function
from .handlers import AlgorithmCallbackHandler
from .stop_condition import AlgorithmStopCondition
from .trust_region import TrustRegion
from .utils import distance_between_tensors, ball_perturb
from .value_normalizer import ValueNormalizer


class ConvergenceAlgorithm:
    def __init__(
        self,
        env: Function,
        curr_point: Tensor,
        function_opt: Optimizer,
        epsilon: float,
        epsilon_factor: float,
        min_epsilon: float,
        perturb: float,
        max_batch_size: int = 1024,
        num_of_batch_reply: int = 32,
        maximum_movement_for_shrink: float = math.inf,
        value_normalizer: ValueNormalizer = None,
        trust_region: TrustRegion = None,
        dtype: torch.dtype = torch.float64,
        device: int = None,
        use_tqdm_bar: bool = True,
        logger: Logger = logging.getLogger(__name__),
    ):
        self.env = env
        self.curr_point = curr_point
        self.best_model = copy.deepcopy(curr_point)
        self.function_opt = function_opt
        self.epsilon = epsilon
        self.min_epsilon = min_epsilon
        self.epsilon_factor = epsilon_factor
        self.perturb = perturb
        self.max_batch_size = max_batch_size
        self.num_of_batch_reply = num_of_batch_reply
        self.maximum_movement_for_shrink = maximum_movement_for_shrink
        self.value_normalizer = value_normalizer
        self.trust_region = trust_region
        self.dtype = dtype
        self.device = device
        self.use_tqdm_bar = use_tqdm_bar
        self.logger = logger

    @property
    def best_point_until_now(self):
        return self.best_model.detach().clone()

    @property
    def curr_point_to_draw(self):
        return self.curr_point.detach()

    def real_data(self, data: Tensor):
        if self.trust_region:
            data = self.env.denormalize(self.trust_region.inverse(data))
        return data

    def eval_data(self, data: Tensor):
        data = self.real_data(data)
        return self.env.evaluate(data)

    def train(
        self,
        epochs: int,
        exploration_size: int,
        num_loop_without_improvement: int,
        min_iteration_before_shrink: int,
        surrogate_model_training_epochs: int = 60,
        warmup_minibatch: int = 5,
        warmup_loops: int = 6,
        stopping_conditions: List[AlgorithmStopCondition] = None,
        callback_handlers: List[AlgorithmCallbackHandler] = None,
        **kwargs,
    ):
        self.logger.info(
            f"Starting running {self.__class__.__name__} for {epochs} epochs"
        )
        stopping_conditions = stopping_conditions or []
        callback_handlers = callback_handlers or []

        for callback_handler in callback_handlers:
            callback_handler.on_algorithm_start(self)

        self.training_start_hook(
            epochs,
            exploration_size,
            num_loop_without_improvement,
            min_iteration_before_shrink,
            surrogate_model_training_epochs,
            warmup_minibatch,
            callback_handlers,
            **kwargs,
        )

        # to prevent error if database is not assigned before exception
        database = torch.tensor([])

        try:
            database, evaluations = self.explore(warmup_minibatch * exploration_size)
            num_of_samples = database.shape[-2]
            batch_size = min(self.max_batch_size, num_of_samples)

            self.warm_up(batch_size, database, evaluations, warmup_loops)

            best_model_value = self.eval_data(self.best_point_until_now)
            reply_memory_size = self.num_of_batch_reply * exploration_size
            no_improvement_in_model_count = 0
            counter = 0
            num_of_shrinks = 0
            last_tr_unreal_best = self.best_point_until_now.clone()
            self.logger.info(f"starting on {best_model_value}")

            epoch_loop = (
                trange(epochs, desc=f"Training EGL {epochs} epochs")
                if self.use_tqdm_bar
                else range(epochs)
            )
            for _ in epoch_loop:
                counter += 1
                # Explore
                samples, new_evaluations = self.explore(exploration_size)
                database = torch.cat((database, samples), dim=(len(samples.shape) - 2))[
                    ..., -reply_memory_size:, :
                ]
                evaluations = torch.cat(
                    (evaluations, new_evaluations), dim=(len(evaluations.shape) - 1)
                )[..., -reply_memory_size:]

                num_of_samples = database.shape[-2]
                batch_size = min(self.max_batch_size, num_of_samples)
                test_loss = self.train_surrogate(database, evaluations, batch_size)
                self.train_model()

                # Handle end of epoch
                for handler in callback_handlers:
                    handler.on_epoch_end(
                        self,
                        database=database,
                        test_losses=test_loss,
                        best_model_value=best_model_value,
                    )

                # Check improvement
                new_model_evaluation = self.eval_data(self.curr_point)
                if best_model_value > new_model_evaluation:
                    self.logger.info(
                        f"Improved best known point to {best_model_value} From {new_model_evaluation} In {self.env}"
                    )
                    self.best_model = copy.deepcopy(self.curr_point)
                    best_model_value = new_model_evaluation
                else:
                    self.logger.warning(
                        f"No improvement ({no_improvement_in_model_count}) for value {new_model_evaluation} "
                        f"in {self.env}"
                    )
                    no_improvement_in_model_count += 1

                # Shrink Trust region
                if (
                    no_improvement_in_model_count >= num_loop_without_improvement
                    and counter >= min_iteration_before_shrink
                ):
                    counter = 0
                    no_improvement_in_model_count = 0
                    num_of_shrinks += 1

                    unreal_distance_between_bests = distance_between_tensors(
                        last_tr_unreal_best,
                        self.best_model.detach().clone(),
                    )
                    self.epsilon *= self.epsilon_factor
                    self.epsilon = max(self.epsilon, self.min_epsilon)
                    should_shrink_in_addition_to_move = (
                        self.maximum_movement_for_shrink > unreal_distance_between_bests
                    )
                    if self.trust_region:
                        self.before_shrinking_hook()
                        best_parameters_real = self.trust_region.inverse(
                            self.best_model.detach().clone()
                        )
                        real_database = self.trust_region.inverse(database.detach())
                        if should_shrink_in_addition_to_move:
                            self.logger.info(
                                f"Shrinking trust region, movement {unreal_distance_between_bests}, new center is "
                                f"{self.env.denormalize(best_parameters_real).tolist()} with {self.trust_region}"
                            )

                            self.trust_region.squeeze(best_parameters_real)
                        else:
                            self.trust_region.move_center(best_parameters_real)
                            self.logger.info(
                                f"Moving trust region after moving {unreal_distance_between_bests}, new center is "
                                f"{self.env.denormalize(best_parameters_real).tolist()} with {self.trust_region}"
                            )

                        reply_memory_size = self.num_of_batch_reply * exploration_size
                        self.curr_point = nn.Parameter(
                            self.trust_region.map(best_parameters_real).clone()
                        )
                        self.best_model = copy.deepcopy(self.curr_point)
                        self.after_shrinking_hook()
                        # NOTE - I reset this network only if trust region has changed
                        #       Because If it has not the network should look the same
                        database = self.trust_region.map(real_database)
                        self.warm_up(batch_size, database, evaluations, warmup_loops)
                    self.logger.info(f"Shrinking sample radius to {self.epsilon}")
                    self.logger.info(f"Space status {self.env}")
                    for handler in callback_handlers:
                        handler.on_algorithm_update(
                            self, database=database, best_model_value=best_model_value
                        )

                for stop_condition in stopping_conditions:
                    if stop_condition.should_stop(self, counter=num_of_shrinks):
                        raise AlgorithmFinished(
                            stop_condition.REASON.format(
                                alg=self.__class__.__name__,
                                env=self.env,
                                best_point=self.best_point_until_now,
                                tr=self.trust_region,
                            )
                        )
        except NoMoreBudgetError as e:
            self.logger.warning("No more Budget", exc_info=e)
        except AlgorithmFinished as e:
            self.logger.info(f"{self.__class__.__name__} Finish stopped {e}")

        for handler in callback_handlers:
            self.logger.info(
                f"Calling upon {handler.on_algorithm_end} finishing convergence"
            )
            if database.numel() != 0:
                handler.on_algorithm_end(self, database=database)
            else:
                handler.on_algorithm_end(self)

    def warm_up(self, batch_size, database, evaluations, warmup_loops):
        for i in range(warmup_loops):
            self.logger.info(f"{i} loop for warmup for {self.__class__.__name__}")
            self.train_surrogate(database, evaluations, batch_size)

    def explore(self, exploration_size: int):
        current_model_parameters = self.curr_point.detach()
        new_model_samples = self.samples_points(
            current_model_parameters, exploration_size
        )

        # Evaluate
        evaluations = self.evaluate_point(new_model_samples).to(device=self.device)

        self.value_normalizer.adapt(evaluations)
        return new_model_samples, evaluations

    def samples_points(self, base_point: Tensor, exploration_size: int):
        self.logger.info(
            f"Exploring new data points. Sampling {exploration_size} points"
        )
        new_model_samples = torch.cat(
            (
                ball_perturb(
                    base_point,
                    self.epsilon,
                    exploration_size - 1,
                    self.dtype,
                    self.device,
                ),
                base_point.reshape(1, -1),
            )
        )
        return new_model_samples

    def evaluate_point(self, new_model_samples: Tensor):
        self.logger.info(
            f"Evaluating {len(new_model_samples)} on env with {self.trust_region}"
        )
        return self.eval_data(new_model_samples)

    def train_surrogate(self, samples: Tensor, samples_value: Tensor, batch_size: int):
        raise NotImplementedError()

    def train_model(self):
        raise NotImplementedError()

    def training_start_hook(self, *args, **kwargs):
        pass

    def before_shrinking_hook(self):
        pass

    def after_shrinking_hook(self):
        pass

    @abc.abstractmethod
    def gradient(self, x) -> Tensor:
        raise NotImplementedError()



from typing import Callable

import torch
from torch import Tensor
from torch.nn import Module


def reset_all_weights(model: Module) -> None:
    """
    refs:
        - https://discuss.pytorch.org/t/how-to-re-set-alll-parameters-in-a-network/20819/6
        - https://stackoverflow.com/questions/63627997/reset-parameters-of-a-neural-network-in-pytorch
        - https://pytorch.org/docs/stable/generated/torch.nn.Module.html
    """

    @torch.no_grad()
    def weight_reset(m: Module):
        reset_parameters = getattr(m, "reset_parameters", None)
        if callable(reset_parameters):
            m.reset_parameters()

    # Applies fn recursively to every submodule see: https://pytorch.org/docs/stable/generated/torch.nn.Module.html
    model.apply(fn=weight_reset)


def distance_between_tensors(
    tensor1: torch.Tensor, tensor2: torch.Tensor
) -> torch.Tensor:
    return torch.norm(tensor1 - tensor2)


def ball_perturb(
    ball_center: Tensor,
    eps: float,
    num_samples: int,
    dtype: torch.dtype = torch.float64,
    device: int = None,
) -> Tensor:
    ball_dim_size = ball_center.shape[-1]

    perturb = (
        torch.FloatTensor(num_samples, ball_dim_size)
        .to(device=device, dtype=dtype)
        .normal_()
    )
    mag = torch.FloatTensor(num_samples, 1).to(device=device, dtype=dtype).uniform_()
    perturb = perturb / (torch.norm(perturb, dim=1, keepdim=True) + 1e-8)

    explore = ball_center + eps * mag * perturb
    return explore


def hessian_from_gradient_network(grad_network, x):
    jacobian = (
        torch.vmap(torch.func.jacrev(grad_network), randomness="different")
        if len(x.shape) == 2
        else torch.func.jacrev(grad_network)
    )
    j = jacobian(x)
    return (j + j.transpose(-2, -1)) / 2


def loss_from_taylor_loss(
    taylor_loss, loss: Callable
) -> Callable[[Tensor, Tensor, Tensor, Tensor], Tensor]:
    def calc_loss(x_i, x_j, y_i, y_j):
        value, target = taylor_loss(x_i, x_j, y_i, y_j)
        return loss(value, target)

    return calc_loss



import torch
from abc import ABC

from torch import Tensor


class ValueNormalizer(ABC):
    def map(self, tensor: Tensor) -> Tensor:
        raise NotImplementedError()

    def inverse(self, tensor: Tensor) -> Tensor:
        raise NotImplementedError()

    def adapt(self, new_data: Tensor):
        pass


class AdaptedOutputUnconstrainedMapping(ValueNormalizer):
    def __init__(
        self,
        outlier: float = 0.1,
        norm_lr: float = 0.1,
        output_epsilon: float = 1e-5,
        squash_eps: float = 1e-5,
        y1: float = -1,
        y2: float = 1,
    ):
        """
        This class is mapping using 2 functions
        1.  linear function that with these points (Q[outlier], y1), (Q[1-outlier], y2)
            Q[outlier] is the outlier percentile of the data
        2.  log function (depends on the output of the first function)
        This class also adapt the linear mapping slowly and give more gravity for new data
        :param epsilon: infinitesimal number to avoid dividing by zero
        """
        self.outlier = outlier
        self.output_epsilon = output_epsilon
        self.norm_lr = norm_lr
        self.squash_eps = squash_eps
        self.y1 = y1
        self.y2 = y2
        self.m = None
        self.n = None

    def adapt(self, new_data):
        data_size = len(new_data)
        outlier = max(int(data_size * self.outlier), 1)

        x2, _ = torch.kthvalue(new_data, data_size - outlier, dim=0)
        x1, _ = torch.kthvalue(new_data, outlier, dim=0)

        m = (self.y2 - self.y1) / (x2 - x1 + self.output_epsilon)
        n = self.y2 - m * x2

        if self.m is None or self.n is None:
            self.m = m
            self.n = n
        else:
            self.m = (1 - self.norm_lr) * self.m + self.norm_lr * m
            self.n = (1 - self.norm_lr) * self.n + self.norm_lr * n

    def map(self, tensor):
        if not self.m or not self.n:
            return tensor
        tensor = tensor.to(device=self.m.device) * self.m + self.n

        x_clamp_up = torch.clamp(tensor, min=self.squash_eps)
        x_clamp_down = torch.clamp(tensor, max=-self.squash_eps)

        x_log_up = torch.log(x_clamp_up) + 1
        x_log_down = -torch.log(-x_clamp_down) - 1
        tensor = (
            x_log_up * (tensor >= 1).float()
            + tensor * (tensor >= -1).float() * (tensor < 1).float()
            + x_log_down * (tensor < -1).float()
        )
        return tensor

    def inverse(self, tensor):
        x_clamp_up = torch.clamp(tensor, min=self.squash_eps)
        x_clamp_down = torch.clamp(tensor, max=-self.squash_eps)

        x_exp_up = torch.exp(x_clamp_up - 1)
        x_exp_down = -torch.exp(-(x_clamp_down + 1))

        tensor = (
            x_exp_up * (tensor >= 1).float()
            + tensor * (tensor >= -1).float() * (tensor < 1).float()
            + x_exp_down * (tensor < -1).float()
        )

        tensor = (tensor - self.n) / self.m
        return tensor



import logging
import math
from logging import Logger
from typing import List

import numpy as np
import torch
from cma import CMAEvolutionStrategy

from .exception import AlgorithmFinished, NoMoreBudgetError
from .function import Function
from .handlers import AlgorithmCallbackHandler
from .stop_condition import AlgorithmStopCondition
from .trust_region import TrustRegion


class TRCovarianceMatrixAdaptation:
    def __init__(
        self,
        env: Function,
        initial_point: np.ndarray,
        trust_region: TrustRegion = None,
        logger: Logger = None,
    ):
        self.curr_best_point = initial_point
        self.env = env
        self.trust_region = trust_region
        self.logger = logger or logging.getLogger(__name__)

    @property
    def real_cma_evaluator(self):
        def evaluator(x):
            return self.env.evaluate(self.trust_region.inverse(x))

        return evaluator

    def train(
        self,
        num_of_epoch_with_no_improvement: int = None,
        stopping_conditions: List[AlgorithmStopCondition] = None,
        callback_handlers: List[AlgorithmCallbackHandler] = None,
    ):
        self.logger.info(f"starting algorithm cma for space {self.env}")
        callback_handlers = callback_handlers or []
        stopping_conditions = stopping_conditions or []

        for handler in callback_handlers:
            handler.on_algorithm_start(self)

        points_found = np.array([])
        cma_es = CMAEvolutionStrategy(self.curr_best_point, 0.5)
        try:
            no_improvement_counter = 0
            curr_value = math.inf
            while True:
                while not cma_es.stop():
                    self.logger.info(f"CMA - new iteration for {self.env}")
                    solutions = np.array(cma_es.ask())
                    solutions_value = (
                        self.real_cma_evaluator(torch.from_numpy(solutions))
                        .cpu()
                        .numpy()
                    )

                    if len(points_found) > 0:
                        concat_solutions = np.concatenate((points_found, solutions))
                    else:
                        concat_solutions = solutions
                    points_found = np.unique(concat_solutions, axis=0)

                    cma_es.tell(solutions, solutions_value.tolist())
                    cma_es.disp()
                    best_idx = np.argmin(solutions_value)
                    best_point = solutions[best_idx]
                    if solutions_value[best_idx] >= curr_value:
                        self.logger.info(
                            f"No improvement, counter {no_improvement_counter}/{num_of_epoch_with_no_improvement}"
                        )
                        no_improvement_counter += 1
                    else:
                        no_improvement_counter = 0
                        self.logger.info(
                            f"improved by {solutions_value[best_idx] - curr_value}"
                        )
                    self.curr_best_point = best_point
                    curr_value = solutions_value[best_idx]
                    for handler in callback_handlers:
                        handler.on_epoch_end(
                            self,
                            database=torch.tensor(solutions),
                        )

                    for stop_condition in stopping_conditions:
                        if stop_condition.should_stop(
                            self, counter=no_improvement_counter
                        ):
                            raise AlgorithmFinished(
                                stop_condition.REASON.format(
                                    alg="CMA",
                                    env=self.env,
                                    best_point=best_point,
                                    tr=self.trust_region,
                                )
                            )
                real_best_point = self.trust_region.inverse(self.curr_best_point)
                self.logger.info(
                    f"Shrinking CMA around {real_best_point} with {self.trust_region}"
                )
                self.trust_region.squeeze(real_best_point)
                self.curr_best_point = self.trust_region.map(real_best_point)
                cma_es = CMAEvolutionStrategy(self.curr_best_point, 0.5)
                for handler in callback_handlers:
                    handler.on_algorithm_update(self)
        except NoMoreBudgetError as e:
            self.logger.warning("Exceeded budget", exc_info=e)
        except AlgorithmFinished as e:
            self.logger.info(f"CMA Finish stopped {e}")
        finally:
            for handler in callback_handlers:
                handler.on_algorithm_end(self)
            if points_found.size:
                return points_found, self.real_cma_evaluator(
                    torch.from_numpy(points_found)
                )
            return np.array([]), np.array([])



from typing import Callable

from torch import Tensor
from torch.nn import Module
from torch.optim import Optimizer
from torch.utils.data import DataLoader, BatchSampler, RandomSampler

from .convergence import ConvergenceAlgorithm
from .datsets import PointDataset


class IGL(ConvergenceAlgorithm):
    def __init__(
        self,
        *args,
        training_epochs: int,
        surrogate_model: Module,
        surrogate_opt: Optimizer,
        loss: Callable,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.training_epochs = training_epochs
        self.surrogate_model = surrogate_model
        self.surrogate_opt = surrogate_opt
        self.loss = loss

    def train_surrogate(self, samples: Tensor, samples_value: Tensor, batch_size: int):
        self.surrogate_model.train()
        mapped_evaluations = self.value_normalizer.map(samples_value)
        dataset = PointDataset(samples, mapped_evaluations)
        for _ in range(self.training_epochs):
            num_of_minibatch = len(samples) // batch_size
            sampler = BatchSampler(
                RandomSampler(range(num_of_minibatch)),
                batch_size=batch_size,
                drop_last=False,
            )
            data_loader = DataLoader(dataset, sampler=sampler)
            for i_index in data_loader:
                x_i = samples[i_index]
                y_i = mapped_evaluations[i_index]

                self.surrogate_opt.zero_grad()
                self.function_opt.zero_grad()
                predicted_value = self.surrogate_model(x_i)

                loss = self.loss(predicted_value.squeeze(), y_i)
                loss.backward()
                self.surrogate_opt.step()
        self.surrogate_model.eval()

    def train_model(self):
        self.function_opt.zero_grad()
        loss = self.surrogate_model(self.curr_point)
        loss.backward()
        self.function_opt.step()
        self.logger.info(
            f"Algorithm {self.__class__.__name__} updated after loss {loss}"
        )



from typing import Callable

import torch
from torch import Tensor
from torch.nn import Module

from .distribution import WeightsDistributionBase
from .utils import hessian_from_gradient_network


class GradientLoss(Module):
    def __init__(self, grad_network):
        super().__init__()
        self.grad_network = grad_network

    def forward(self, x_i, x_j, y_i, y_j):
        assert len(x_i.shape) <= 2, "Cant handle multiple batches yet"
        grad_x_i = self.grad_network(x_i)

        x_delta = x_j - x_i
        value = (x_delta * grad_x_i).sum(dim=1) + self.taylor_remainder(x_delta)
        target = y_j - y_i

        return value, target

    def taylor_remainder(self, x_delta):
        return 0


class NaturalHessianLoss(GradientLoss):
    def calculate_hessian(self, x):
        return hessian_from_gradient_network(self.grad_network, x)

    def taylor_remainder(self, x_delta):
        hessian = self.calculate_hessian(x_delta)
        return torch.bmm(
            torch.bmm(
                x_delta.reshape((x_delta.shape[0], 1, x_delta.shape[1])), hessian
            ),
            x_delta.reshape((x_delta.shape[0], x_delta.shape[1], 1)),
        ).squeeze()


class DetachedHessianLoss(NaturalHessianLoss):
    def calculate_hessian(self, x):
        with torch.no_grad():
            return super().calculate_hessian(x).detach()


def loss_with_quantile(
    value: Tensor,
    target: Tensor,
    weights_creator: WeightsDistributionBase,
    loss: Callable,
) -> Tensor:
    smallest_element = (value - target).abs().clone().detach()
    weights = weights_creator.distribute_weights(smallest_element)

    loss = loss(value, target)
    loss = (loss * weights).mean()
    return loss



import abc
from abc import ABC


class AlgorithmStopCondition(ABC):
    REASON: str = ""

    @abc.abstractmethod
    def should_stop(self, alg, **kwargs) -> bool:
        raise NotImplementedError()
